#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <fcntl.h>
#include <unistd.h>
#include <conio.h>
#include <string.h>

#define GLOBAL

#include "config.h"
#include "global.h"
#include "bmpeg.h"

#define SAFE_free(p) if(p) { free(p); p = NULL; }


char mpegfilename[256];
int lastGetSeqHeader = 0;
int MPG_Framenum = 0;
char emptyStr[] = " ";
bool firstOpen = true;
bool MpegOpened = false;
unsigned char *lastframe[3];

static void Clear_Options()
{
  Verbose_Flag = 0;
  Output_Type = 0;
  hiQdither  = 0;
  Output_Type = 9;
  Output_Picture_Filename = emptyStr;
  Frame_Store_Flag = 1; /* to avoid calling dither() twice */
  Display_Progressive_Flag = 1;
  Spatial_Flag = 0;
  Lower_Layer_Picture_Filename = emptyStr;
  Reference_IDCT_Flag = 0;
  Trace_Flag = 0;
  Quiet_Flag = 0;
  Ersatz_Flag = 0;
  Substitute_Picture_Filename  = emptyStr;
  Two_Streams = 0;
  Enhancement_Layer_Bitstream_Filename = emptyStr;
  Big_Picture_Flag = 0;
  Main_Bitstream_Flag = 0;
  Main_Bitstream_Filename = emptyStr;
  Verify_Flag = 0;
  Stats_Flag  = 0;
  User_Data_Flag = 0;
}


void Error(char *text)
{
  fprintf(stderr,text);
//  exit(1);
}

/* return when end of sequence (0) or picture
   header has been parsed (1) */

int Get_SeqHeaders()
{
  ld = &base;
  return Get_Hdr();
}

bool Initialize_Sequence()
{
  int cc, size;
  static int Table_6_20[3] = {6,8,12};

  /* force MPEG-1 parameters for proper decoder behavior */
  /* see ISO/IEC 13818-2 section D.9.14 */
  if (!base.MPEG2_Flag)
  {
    progressive_sequence = 1;
    progressive_frame = 1;
    picture_structure = FRAME_PICTURE;
    frame_pred_frame_dct = 1;
    chroma_format = CHROMA420;
    matrix_coefficients = 5;
  }

  /* round to nearest multiple of coded macroblocks */
  /* ISO/IEC 13818-2 section 6.3.3 sequence_header() */
  mb_width = (horizontal_size+15)/16;
  mb_height = (base.MPEG2_Flag && !progressive_sequence) ? 2*((vertical_size+31)/32)
                                        : (vertical_size+15)/16;

  Coded_Picture_Width = 16*mb_width;
  Coded_Picture_Height = 16*mb_height;

  /* ISO/IEC 13818-2 sections 6.1.1.8, 6.1.1.9, and 6.1.1.10 */
  Chroma_Width = (chroma_format==CHROMA444) ? Coded_Picture_Width
                                           : Coded_Picture_Width>>1;
  Chroma_Height = (chroma_format!=CHROMA420) ? Coded_Picture_Height
                                            : Coded_Picture_Height>>1;
  
  /* derived based on Table 6-20 in ISO/IEC 13818-2 section 6.3.17 */
  block_count = Table_6_20[chroma_format-1];
  
  for (cc=0; cc<3; cc++)
  {
    if (cc==0)
      size = Coded_Picture_Width*Coded_Picture_Height;
    else
      size = Chroma_Width*Chroma_Height;

    if (!(backward_reference_frame[cc] = (unsigned char *)malloc(size)))
      return false;
    if (!(forward_reference_frame[cc] = (unsigned char *)malloc(size)))
      return false;

    if (!(auxframe[cc] = (unsigned char *)malloc(size)))
      return false;

    if(Ersatz_Flag)
      if (!(substitute_frame[cc] = (unsigned char *)malloc(size)))
         return false;

    if (base.scalable_mode==SC_SPAT)
    {
      /* this assumes lower layer is 4:2:0 */
      if (!(llframe0[cc] = (unsigned char *)malloc((lower_layer_prediction_horizontal_size*lower_layer_prediction_vertical_size)/(cc?4:1))))
        return false;
      if (!(llframe1[cc] = (unsigned char *)malloc((lower_layer_prediction_horizontal_size*lower_layer_prediction_vertical_size)/(cc?4:1))))
        return false;
    }
  }

  /* SCALABILITY: Spatial */
  if (base.scalable_mode==SC_SPAT)
  {
    if (!(lltmp = (short *)malloc(lower_layer_prediction_horizontal_size*((lower_layer_prediction_vertical_size*vertical_subsampling_factor_n)/vertical_subsampling_factor_m)*sizeof(short))))
      return false;
  }
  if(firstOpen) {
    /* Clip table */
    if (!(Clip=(unsigned char *)malloc(1024)))
    {
      //Error("Clip[] malloc failed\n");
      return false;
    }

    Clip += 384;

    for (int i=-384; i<640; i++)
      Clip[i] = (i<0) ? 0 : ((i>255) ? 255 : i);
    firstOpen = false;
  }
  
  return true;
}

static int isTransportStream(void)
{
	int syncPos[1024];
	int numSync = 0;
	int i = 0, j, k;

	while (i < 1024)
	{
		if(Show_Bits(8)==0x47)
		{
			syncPos[numSync] = i;
			numSync++;
			for (j=0; j < numSync -1; j++)
			{
				for (k=j+1; k < numSync; k++)
				if ((syncPos[k] - syncPos[j]) == 188)
				{
					/*seek to the beginning of the transport stream block*/
      				lseek(base.Infile, (long)syncPos[j], 0);
					return 1;
				}
			}
		}
		Get_Bits(8);
		i++;
	}
	return 0;
}

static void Deinitialize_Sequence()
{
  int i;

  /* clear flags */
  base.MPEG2_Flag=0;

  for(i=0;i<3;i++)
  {
    SAFE_free(backward_reference_frame[i]);
    SAFE_free(forward_reference_frame[i]);
    SAFE_free(auxframe[i]);

    if (base.scalable_mode==SC_SPAT)
    {
     SAFE_free(llframe0[i]);
     SAFE_free(llframe1[i]);
    }
  }

  if (base.scalable_mode==SC_SPAT)
    SAFE_free(lltmp);

  lastGetSeqHeader = 0;
  MpegOpened = false;
  MPG_Framenum = 0;
}

bool GetCurMpegInfo(SMpegInfo *MpgInf)
{
  if(MpegOpened) {
     MpgInf->frame_rate = (float)frame_rate;
     MpgInf->width = Coded_Picture_Width;
     MpgInf->height = Coded_Picture_Height;
     if (chroma_format == CHROMA420)
       MpgInf->chroma_format = 0;
     else if (chroma_format == CHROMA422)
         MpgInf->chroma_format = 1;
     else if (chroma_format == CHROMA444)
         MpgInf->chroma_format = 2;
     return true;
  } else {
     MpgInf->frame_rate = 0.0f;
     MpgInf->width = 0;
     MpgInf->height = 0;
     MpgInf->chroma_format = -1;
     return false;
  }
}


bool OpenMPEG(const char *FName, SMpegInfo *MpgInf)
{
  int ret, code, i;
  int Transport_Stream_Flag = 0;

  Clear_Options();
  ld = &base; /* select base layer context */
  strcpy(mpegfilename, FName);
  Main_Bitstream_Filename = mpegfilename;
  
  /* open MPEG base layer bitstream file(s) */
  /* NOTE: this is either a base layer stream or a spatial enhancement stream */
  if ((base.Infile=open(Main_Bitstream_Filename,O_RDONLY|O_BINARY))<0)
  {
    fprintf(stderr,"Base layer input file %s not found\n", Main_Bitstream_Filename);
    return false;
  }


  if(base.Infile != 0)
  {
    Initialize_Buffer(); 
 
	if (isTransportStream())
	{
	  Transport_Stream_Flag = 1;
//	  pesPID = -1;
          Initialize_Buffer();
          close(base.Infile);
          return false;
	}
	else
	{
		lseek(base.Infile, 0l, 0);
		Initialize_Buffer(); 
	}

	if (!Transport_Stream_Flag)
	{
    	next_start_code();		/* wait for the 3-byte 0x000001 value is in buf */
 	   	code = Get_Bits(32);
	   	if ((code & VIDEO_ELEMENTARY_STREAM) == VIDEO_ELEMENTARY_STREAM)
	   		 code = VIDEO_ELEMENTARY_STREAM;
		switch(code)
		{
		case SEQUENCE_HEADER_CODE:
			fprintf(stderr, "Start stream type 0x%x\n", code);
			break;
		case PACK_START_CODE:				/* 0x000001BA */
			System_Stream_Flag = 1;
			fprintf(stderr, "Start stream type 0x%x\n", code);
			break;
		case VIDEO_ELEMENTARY_STREAM:		/* 0x000001E0  -  0x000001EF*/
			fprintf(stderr, "Start stream type 0x%x\n", code);
			code = Get_Bits(16);
			if (code > 0)
				System_Stream_Flag = 1;
			break;
		default:
			fprintf(stderr, "Unable to recognize stream type 0x%x\n", code);
			exit(0);
			break;
		}
		lseek(base.Infile, 0L, 0);
		Initialize_Buffer(); 
	}
  }
//  lseek(base.Infile, 0l, 0);
//  Initialize_Buffer();

//  Initialize_Reference_IDCT();
  Initialize_Fast_IDCT();
#ifdef VERIFY
  Clear_Verify_Headers();
#endif /* VERIFY */

  if ((lastGetSeqHeader=Get_SeqHeaders())!=0 && Initialize_Sequence()) {

     MpegOpened = true;
     MPG_Framenum = 0;
     GetCurMpegInfo(MpgInf);
//     lseek(base.Infile, 0l, 0);
//     Initialize_Buffer();
     return true;
  }

  Main_Bitstream_Filename = emptyStr;
  close(base.Infile);
  return false;
}

//current_frame

bool DecodeNextMpegFrame(int FramesToDrop)
{
   for(int i=0;i<FramesToDrop && lastGetSeqHeader!=0;i++)
     lastGetSeqHeader=Get_SeqHeaders();
   if(lastGetSeqHeader != 0) {
     Decode_Picture(MPG_Framenum, MPG_Framenum);
//     Output_Last_Frame_of_Sequence(MPG_Framenum);
     MPG_Framenum++;
    if ((picture_structure==FRAME_PICTURE || Second_Field) && (picture_coding_type==B_TYPE))
      memcpy(lastframe, auxframe, sizeof(unsigned char*)*3);
    else
      memcpy(lastframe, forward_reference_frame, sizeof(unsigned char*)*3);
      
     lastGetSeqHeader=Get_SeqHeaders();
     return true;
   }
   return false;
}

void CloseMPEG()
{
   if(MpegOpened) {
     Initialize_Buffer();
     Flush_Buffer32();
     Deinitialize_Sequence();
#ifdef VERIFY
    Clear_Verify_Headers();
#endif /* VERIFY */
     MPG_Framenum = 0;
     MpegOpened = false;
     Main_Bitstream_Filename = emptyStr;
     close(base.Infile);
     bzero(&base, sizeof(base));
     ld = NULL;
//     base.Infile = NULL;
   }
}
/*
int main(int argc, char **argv)
{
  SMpegInfo myMpegIng;
//  if(argc<2) {
    if(OpenMPEG("ifa.mpg", &myMpegIng))
    {
      printf("succeed !!! :=) openening mpeg !!!\n");
      printf("frame rate = %f\n", myMpegIng.frame_rate);
      printf("size = %ix%i\n", myMpegIng.width, myMpegIng.height);
      printf("chrome format = %i\n", myMpegIng.chroma_format);
      printf("progessive seq = %i\n",progressive_sequence);
      getch();
    }
    else {
      printf("failed openening mpeg :( !!!\n");
      return -1;
    }
  for(;;) {
    if(!DecodeNextMpegFrame(0))
      break;
    printf("frames %i OK - ",MPG_Framenum);
    printf(" picture_structure %i, picture_coding_type %i, Second_Field %i\n",picture_structure, picture_coding_type, Second_Field);
    if (Two_Streams)
      printf("Two streams !! OK\n");
  }
  printf("frames %i OK\n",MPG_Framenum);
  printf("press key to close!!!!!!!!!!!!!!\n");
  getch();
  CloseMPEG();
  printf("press key !!!!!!!!!!!!!!\n");
  getch();
    
//  }
//  else {
    if(OpenMPEG("first2.mpg", &myMpegIng))
    {
      printf("succeed !!! :=) openening mpeg !!!\n");
      printf("frame rate = %f\n", myMpegIng.frame_rate);
      printf("size = %ix%i\n", myMpegIng.width, myMpegIng.height);
      printf("chrome format = %i\n", myMpegIng.chroma_format);
      printf("progessive seq = %i\n",progressive_sequence);
      getch();
    }
    else {
      printf("failed openening mpeg :( !!! %s\n",argv[1]);
      return -1;
    }
//  }
  //getch();
  for(;;) {
    if(!DecodeNextMpegFrame(0))
      break;
    printf("frames %i OK -",MPG_Framenum);
    printf(" picture_structure %i, picture_coding_type %i, Second_Field %i\n",picture_structure, picture_coding_type, Second_Field);
    if (Two_Streams)
      printf("Two streams !! OK\n");
  }
  printf("frames %i OK\n",MPG_Framenum);


  CloseMPEG();
  return 0;
}*/
